﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.ComponentModel;
using System.Web;

namespace BendayObjectDataSource.Common
{
    public abstract class BendayDataSourceControl<T> : DataSourceControl, IDataSource, IListSource  where T : class
    {
        public BendayDataSourceControl()
        {
            
        }

        protected override void OnLoad(EventArgs e)
        {
            this.Page.PreRender += new EventHandler(Page_PreRender);

            base.OnLoad(e);
        }

        /// <summary>
        /// This specifies where the data source is stored between web requests.
        /// </summary>
        public StorageLocationType DataSourceStorageLocation { get; set; }
        
        /// <summary>
        /// If the DataSourceStorageLocation is set to Session, this is the name of the key
        /// in the ASP.NET session state.  
        /// </summary>
        public string DataSourceSessionKeyName { get; set; }

        void Page_PreRender(object sender, EventArgs e)
        {
            if (m_dataSource != null && DataSourceStorageLocation == StorageLocationType.ViewState)
            {
                SerializeDataSourceToViewState();
            }
            else if (m_dataSource != null && DataSourceStorageLocation == StorageLocationType.Session)
            {
                SerializeDataSourceToSession();                
            }
        }

        private void CheckDataSourceIsPopulated()
        {
            if (DataSourceStorageLocation == StorageLocationType.ViewState)
            {
                CheckDataSourceIsPopulatedFromViewState();    
            }
            else
            {
                CheckDataSourceIsPopulatedFromSession();
            }
            
        }

        private void CheckDataSourceIsPopulatedFromSession()
        {
            if (m_dataSource == null)
            {
                IList<T> temp = HttpContext.Current.Session[DataSourceSessionKeyName] as IList<T>;

                if (temp == null)
                {
                    temp = new BindingList<T>();
                }

                m_dataSource = temp;
            }
        }

        private void CheckDataSourceIsPopulatedFromViewState()
        {
            if (m_dataSource == null && ViewState != null)
            {
                object temp = ViewState["dataSource"];

                if (temp != null)
                {
                    byte[] dataSourceAsByteArray = ViewState["dataSource"] as byte[];

                    if (dataSourceAsByteArray != null)
                    {
                        MemoryStream stream = new MemoryStream(dataSourceAsByteArray);

                        BinaryFormatter formatter = new BinaryFormatter();

                        m_dataSource = formatter.Deserialize(stream) as IList<T>;
                    }
                }
                else
                {
                    m_dataSource = new BindingList<T>();
                }
            }
        }
        private IList<T> m_dataSource;
        public IList<T> DataSource
        {
            get
            {
                CheckDataSourceIsPopulated();

                return m_dataSource;
            }
            set
            {
                if (value == null)
                {
                    m_dataSource = new BindingList<T>();
                }
                else
                {
                    m_dataSource = new BindingList<T>(value);
                }
            }
        }
        
        public abstract string IdPropertyName
        {
            get;
        }

        public abstract T FindById(int id);

        public abstract T CreateNew();

        /// <summary>
        /// Remove items by id from the datasource 
        /// </summary>
        /// <param name="id"></param>
        /// <returns>Number of items removed.</returns>
        public int RemoveById(int id)
        {
            if (this.DataSource != null)
            {
                T item = FindById(id);

                if (item != null)
                {
                    this.DataSource.Remove(item);
                    Refresh();
                    return 1;
                }
            }

            return 0;
        }

        private void SerializeDataSourceToSession()
        {
            if (String.IsNullOrEmpty(DataSourceSessionKeyName) == true)
            {
                DataSourceSessionKeyName = Guid.NewGuid().ToString();
            }

            HttpContext.Current.Session[DataSourceSessionKeyName] = m_dataSource;
        }

        private void SerializeDataSourceToViewState()
        {
            BinaryFormatter formatter = new BinaryFormatter();

            MemoryStream stream = new MemoryStream();

            formatter.Serialize(stream, m_dataSource);

            ViewState["dataSource"] = stream.ToArray();
        }

        private BendayDataSourceView<T> m_defaultView;

        protected override DataSourceView GetView(string viewName)
        {
            if (m_defaultView == null)
            {
                m_defaultView = new BendayDataSourceView<T>(this, "DefaultView");
            }

            return m_defaultView;
        }

        public void Refresh()
        {
            this.RaiseDataSourceChangedEvent(new EventArgs());
        }

        public int GetIndexOfSelectedItem(GridView grid)
        {
            if (grid == null || grid.SelectedRow == null)
            {
                return 0;
            }
            else
            {
                return grid.SelectedRow.DataItemIndex;
            }
        }

        public T GetDataItem(GridView grid)
        {
            if (grid == null || grid.SelectedRow == null)
            {
                return default(T);
            }
            else
            {
                return GetDataItem(grid.SelectedRow);
            }
        }

        private T GetDataItem(GridViewRow row)
        {
            if (row == null)
            {
                return default(T);
            }
            else
            {
                if (this.DataSource == null || row.DataItemIndex >= this.DataSource.Count)
                {
                    return default(T);
                }
                else
                {
                    return this.DataSource[row.DataItemIndex];
                }
            }
        }

        public void Add(T addThis)
        {
            if (addThis != null)
            {
                this.DataSource.Add(addThis);
            }
        }
    }
}
