using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web.UI;
using System.Collections;
using System.Diagnostics;

namespace BendayObjectDataSource.Common
{
    public class BendayDataSourceView<T> : DataSourceView where T : class
    {
        private BendayDataSourceControl<T> m_owner;

        public BendayDataSourceView(BendayDataSourceControl<T> owner, string viewName)
            : base(owner, viewName)
        {
            m_owner = owner;            
        }

        private int GetRowCount(IList<T> dataSource)
        {
            if (dataSource == null)
            {
                return 0;
            }
            else
            {
                return dataSource.Count;
            }
        }

        protected override IEnumerable ExecuteSelect(DataSourceSelectArguments arguments)
        {
            arguments.RaiseUnsupportedCapabilitiesError(this);

            arguments.TotalRowCount = GetRowCount(m_owner.DataSource);

            if (String.IsNullOrEmpty(arguments.SortExpression) == true)
            {
                return SelectPagedData(m_owner.DataSource, arguments);    
            }
            else
            {
                return SelectPagedData(
                    DataSourceUtility.Sort<T>(m_owner.DataSource, DataSourceUtility.ParseSortArguments(arguments.SortExpression)), 
                    arguments);                
            }
        }

        private void WriteDictionary(IDictionary keys, string name)
        {
            Debug.WriteLine(String.Format("{0}.Count = {1}", name, keys.Count));

            foreach (DictionaryEntry val in keys)
            {
                Debug.WriteLine(String.Format("{0} -- {1} = {2}", name, val.Key, val.Value));
            }
        }

        private bool HasFieldName(IDictionary values, string fieldName)
        {
            var entry = GetFieldByName(values, fieldName);

            if (entry.Key != null && entry.Key.ToString() == fieldName)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        private DictionaryEntry GetFieldByName(IDictionary values, string fieldName)
        {
            WriteDictionary(values, "GetFieldByName()");
            if (values == null || fieldName == null)
            {
                // DictionaryEntry is non-nullable so I have to return a value.  
                return new DictionaryEntry();
            }

            DictionaryEntry entry = (from e in values.Cast<DictionaryEntry>().AsQueryable<DictionaryEntry>()
                                     where e.Key != null && e.Key.ToString() == fieldName
                                     select e).FirstOrDefault();

            return entry;
        }

        private void UpdateField(T updateThis, string fieldName, object value)
        {
            PropertyReflector reflector = new PropertyReflector(typeof(T), fieldName);

            reflector.SetValue(updateThis, SafeToString(value));
        }

        private string SafeToString(object value)
        {
            if (value == null)
            {
                return String.Empty;
            }
            else
            {
                return value.ToString();
            }
        }

        public override void Update(IDictionary keys, IDictionary values, IDictionary oldValues, DataSourceViewOperationCallback callback)
        {
            if (HasFieldName(keys, m_owner.IdPropertyName) == true)
            {
                var entry = GetFieldByName(keys, m_owner.IdPropertyName);

                int id = Convert.ToInt32(entry.Value);

                T updateThis = m_owner.FindById(id);

                UpdateInstanceFromValues(values, updateThis);

                callback(1, null);

                m_owner.Refresh();
            }
        }

        private void UpdateInstanceFromValues(IDictionary values, T updateThis)
        {
            if (updateThis != null && values != null)
            {
                foreach (DictionaryEntry field in values)
                {
                    UpdateField(updateThis, field.Key.ToString(), field.Value);
                }
            }
        }

        protected override int ExecuteInsert(IDictionary values)
        {
            return base.ExecuteInsert(values);
        }

        public override void Insert(IDictionary values, DataSourceViewOperationCallback callback)
        {
            T instance = m_owner.CreateNew();

            UpdateInstanceFromValues(values, instance);

            m_owner.DataSource.Add(instance);

            callback(1, null);

            m_owner.Refresh();
        }

        private IEnumerable SelectPagedData(IList<T> data, DataSourceSelectArguments arguments)
        {
            int startingRow = 0;

            if (arguments.StartRowIndex >= 0)
            {
                startingRow = arguments.StartRowIndex;
            }

            IEnumerable<T> temp = (from row in data
                                   select row).Skip(startingRow).Take(arguments.MaximumRows);

            return temp;
        }

        public override bool CanUpdate
        {
            get
            {
                return true;
            }
        }

        public override bool CanInsert
        {
            get
            {
                return true;
            }
        }

        public override bool CanRetrieveTotalRowCount
        {
            get
            {
                return true;
            }
        }

        public override bool CanDelete
        {
            get
            {
                return true;
            }
        }

        public override bool CanSort
        {
            get
            {
                return true;
            }
        }

        public override bool CanPage
        {
            get
            {
                return true;
            }
        }

        protected override int ExecuteDelete(IDictionary keys, IDictionary oldValues)
        {
            WriteDictionary(keys, "ExecuteDelete()");
            if (HasFieldName(keys, m_owner.IdPropertyName) == true)
            {
                int id = Convert.ToInt32(GetFieldByName(keys, m_owner.IdPropertyName).Value);

                m_owner.RemoveById(id);
            }

            return 0;
        }
    }
}
