﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace BendayObjectDataSource.Common
{
    public class DataSourceUtility
    {
        public static IList<T> Sort<T>(IList<T> unsorted, SortArgument argument)
        {
            PropertyReflector reflector = new PropertyReflector(typeof(T), argument.SortBy);

            if (reflector.PropertyType == typeof(int))
            {
                return SortByInt32<T>(unsorted, argument, reflector);
            }
            else if (reflector.PropertyType == typeof(float))
            {
                return SortByFloat<T>(unsorted, argument, reflector);
            }
            else if (reflector.PropertyType == typeof(DateTime))
            {
                return SortByDateTime<T>(unsorted, argument, reflector);
            }
            else
            {
                return SortByString<T>(unsorted, argument, reflector);
            }            
        }

        public static string PadDateTime(DateTime value)
        {
            return PadFloat(value.Ticks);
        }

        public static string PadFloat(double value)
        {
            long floatAsLong = Convert.ToInt64(Math.Truncate(value));

            double remainder = value - ((double)floatAsLong);

            string remainderAsString = remainder.ToString();

            int decimalPointPosition = remainderAsString.IndexOf(".");

            if (decimalPointPosition == -1)
            {
                remainderAsString = ".0";
            }
            else
            {
                remainderAsString = remainderAsString.Substring(decimalPointPosition, remainderAsString.Length - decimalPointPosition);    
            }            

            return PadInt64(floatAsLong) + remainderAsString;
        }

        public static string PadInt64(long value)
        {
            int paddedLength = Int64.MaxValue.ToString().Length;

            int valueLength = value.ToString().Length;

            int requiredPadding = paddedLength - valueLength;

            StringBuilder padded = new StringBuilder();

            for (int i = 0; i < requiredPadding; i++)
            {
                padded.Append("0");
            }

            padded.Append(value);

            return padded.ToString();
        }

        public static string PadInt32(int value)
        {
            int paddedLength = Int32.MaxValue.ToString().Length;
            
            int valueLength = value.ToString().Length;

            int requiredPadding = paddedLength - valueLength;

            StringBuilder padded = new StringBuilder();

            for (int i = 0; i < requiredPadding; i++)
            {
                padded.Append("0");
            }

            padded.Append(value);

            return padded.ToString();
        }

        private static IList<T> SortByString<T>(IList<T> unsorted, SortArgument argument, PropertyReflector reflector)
        {
            SortedList<string, T> sorted = new SortedList<string, T>();
            
            string valueToSortOn;

            foreach (T item in unsorted)
            {
                valueToSortOn = reflector.GetValueAsString(item);

                if (sorted.ContainsKey(valueToSortOn) == false)
                {
                    sorted.Add(valueToSortOn, item);
                }
                else
                {
                    sorted.Add(GetUniqueKey(sorted.Keys, valueToSortOn), item);
                }
            }

            if (argument.Direction == SortDirection.Ascending)
            {
                return sorted.Values;
            }
            else
            {
                return sorted.Values.Reverse().ToList();
            }
        }

        private static IList<T> SortByDateTime<T>(IList<T> unsorted, SortArgument argument, PropertyReflector reflector)
        {
            SortedList<string, T> sorted = new SortedList<string, T>();

            string valueToSortOn;

            foreach (T item in unsorted)
            {
                valueToSortOn = PadDateTime(reflector.GetValueAsDateTime(item));

                if (sorted.ContainsKey(valueToSortOn) == false)
                {
                    sorted.Add(valueToSortOn, item);
                }
                else
                {
                    sorted.Add(GetUniqueKey(sorted.Keys, valueToSortOn), item);
                }
            }

            if (argument.Direction == SortDirection.Ascending)
            {
                return sorted.Values;
            }
            else
            {
                return sorted.Values.Reverse().ToList();
            }
        }

        private static IList<T> SortByFloat<T>(IList<T> unsorted, SortArgument argument, PropertyReflector reflector)
        {
            SortedList<string, T> sorted = new SortedList<string, T>();

            string valueToSortOn;

            foreach (T item in unsorted)
            {
                valueToSortOn = PadFloat(reflector.GetValueAsFloat(item));

                if (sorted.ContainsKey(valueToSortOn) == false)
                {
                    sorted.Add(valueToSortOn, item);
                }
                else
                {
                    sorted.Add(GetUniqueKey(sorted.Keys, valueToSortOn), item);
                }
            }

            if (argument.Direction == SortDirection.Ascending)
            {
                return sorted.Values;
            }
            else
            {
                return sorted.Values.Reverse().ToList();
            }
        }

        private static IList<T> SortByInt32<T>(IList<T> unsorted, SortArgument argument, PropertyReflector reflector)
        {
            SortedList<string, T> sorted = new SortedList<string, T>();

            string valueToSortOn;

            foreach (T item in unsorted)
            {
                valueToSortOn = PadInt32(reflector.GetValueAsInt32(item));

                if (sorted.ContainsKey(valueToSortOn) == false)
                {
                    sorted.Add(valueToSortOn, item);
                }
                else
                {
                    sorted.Add(GetUniqueKey(sorted.Keys, valueToSortOn), item);
                }
            }

            if (argument.Direction == SortDirection.Ascending)
            {
                return sorted.Values;
            }
            else
            {
                return sorted.Values.Reverse().ToList();
            }
        }

        public static string GetUniqueKey(IList<string> alreadyInUseKeys, string proposedKey)
        {
            IList<string> toLowerKeys = new List<string>();

            foreach (string item in alreadyInUseKeys)
            {
                toLowerKeys.Add(item.ToLower());
            }

            if (toLowerKeys.Contains(proposedKey.ToLower()) == false)
            {
                return proposedKey;
            }
            else
            {
                string currentAttempt = proposedKey;
                int uniquifier = 1;

                while (toLowerKeys.Contains(currentAttempt.ToLower()) == true)
                {
                    currentAttempt = String.Format("{0}_{1}", proposedKey, uniquifier);
                    uniquifier++;
                }

                return currentAttempt;
            }
        }

        public static SortArgument ParseSortArguments(string sortArgumentString)
        {
            if (sortArgumentString == null)
            {
                return null;
            }

            string trimmed = sortArgumentString.Trim();

            SortDirection direction = SortDirection.Ascending;

            if (trimmed.ToLower().EndsWith(" desc") == true)
            {
                direction = SortDirection.Descending;
                trimmed = trimmed.Remove(trimmed.Length - 5, 5);
            }
            else if (trimmed.ToLower().EndsWith(" asc") == true)
            {
                direction = SortDirection.Ascending;
                trimmed = trimmed.Remove(trimmed.Length - 4, 4);
            }

            SortArgument arg = new SortArgument(trimmed, direction);

            return arg;
        }
    }
}
